package com.hemmis.sudoku;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;


/**
 * Represents a sudoku field of 9 by 9
 * @author sst
 */
public class Field {
	public static final int SIZE = 9;
	private Cell cells[][];
	private List<Group> rows=new ArrayList<Group>();
	private List<Group> columns=new ArrayList<Group>();
	private List<Group> squares=new ArrayList<Group>();
	private Set<Group> allGroups=new HashSet<Group>();
	
	
	public Field() {
		cells = new Cell[SIZE][];
		
		for (int row=0;row <SIZE; row ++) {
			cells[row] = new Cell[SIZE];
			for (int col=0; col<SIZE; col++) {
				cells[row][col] = new Cell(row,col);
			}
		}
		
		for (int row=0; row<SIZE; row++) {
			Group group = new Group(Arrays.asList(cells[row]));
			rows.add(group);
			allGroups.add(group);
		}
		
		for (int col=0; col<SIZE; col++) {
			List<Cell> colCells = new ArrayList<Cell>(SIZE);
			for (int row=0; row<SIZE; row++) {
				colCells.add(getCell(row,col));
			}
			Group group = new Group(colCells);
			columns.add(group);
			allGroups.add(group);
		}
		
		for (int squareRow = 0;squareRow<3; squareRow++) {
			for (int squareCol = 0;squareCol<3; squareCol++) {
				Group group = getSquare(squareRow, squareCol);
				squares.add(group);
				allGroups.add(group);
			}
		}
	}
	
	/**
	 * Get a Group representing a 3x3 cell
	 * @param squareRow
	 * @param squareCol
	 * @return
	 */
	private Group getSquare(int squareRow, int squareCol) {
		List<Cell> squareCells = new ArrayList<Cell>(SIZE);
		for (int row=0; row<3; row++) {
			for (int col=0; col<3; col++) {
				squareCells.add(getCell(squareRow*3+row,squareCol*3+col));
			}
		}
		return new Group(squareCells);
	}
	
	private Cell getCell(int row, int col) {
		return cells[row][col];
	}
	
	public boolean isKnown() {
		for (Group row:rows) {
			if (!row.isKnown()) {
				return false;
			}
		}
		return true;
	}
	
	
	public BigInteger getNrOfPossibilities () {
		BigInteger nrOfPossibilities = new BigInteger("1");
		for (Group row:rows) {
			nrOfPossibilities = nrOfPossibilities.multiply(row.getNrOfPossibilities()); 
		}
		return nrOfPossibilities;
	}
	
	public void solve() throws NoMoreValuesException {
		BigInteger nrOfPossibilities = getNrOfPossibilities();
		BigInteger oldNrOfPossiblities = null;
		System.out.println("Nr of possiblities = "+nrOfPossibilities);
		while (nrOfPossibilities.compareTo(new BigInteger("1"))>0 && (oldNrOfPossiblities==null || nrOfPossibilities.compareTo(oldNrOfPossiblities)<0)) {
			for (Group group:allGroups) {
				group.solve();
			}
			oldNrOfPossiblities = nrOfPossibilities;
			nrOfPossibilities = getNrOfPossibilities();
			System.out.println("Nr of possiblities = "+nrOfPossibilities);
		}
	}
	
	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer();
		for (Group row: rows) {
			sb.append(row);
			sb.append("\n");
		}
		return sb.toString();
	}
	
	public static Field fromFile(File file) {
		try {
			Field field = new Field();
			BufferedReader br=new BufferedReader(new FileReader(file));
			String line;
			int row = 0;
			while (row <9 && null!=(line = br.readLine())) {
				int col = 0;
				String [] values;
				if (line.contains(",")) {
					values = line.split(",");
				} else {
					values = line.split("\\t");
				}
				for (String valueString : values) {
					if (valueString.trim().length()>0) {
						Value value = Value.fromString(valueString.trim());
						if (value!=null) {
							field.getCell(row, col).setValue(value);
						}
					}
					col++;
				}
				row++;	
			}
			br.close();
			return field;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return null;
	}
}
