package com.hemmis.sudoku;

import java.math.BigInteger;
import java.util.List;


/**
 * Group of cells in a sudoku field. Can be a row/colum/3x3 block
 * @author sst
 *
 */
public class Group {
	private List<Cell> cells;
	/**
	 * Constructor
	 * @param cells the cells from the group
	 */
	public Group(List<Cell> cells) {
		this.cells = cells;
	}
	/**
	 * Check if all cells in the group are known
	 * @return
	 */
	public boolean isKnown() {
		for (Cell cell:cells) {
			if (!cell.isKnown()) {
				return false;
			}
		}
		return true;
	}
	
	/**
	 * Try to solve cells in a group
	 * @return if the group is completely solved/known
	 * @throws NoMoreValuesException
	 */
	public boolean solve() throws NoMoreValuesException {
		if (isKnown()) {
			return true;
		}
		removeDuplicates();
		selectUniquePossibilities();
		removeGroups();
		return isKnown();
	}
	
	/**
	 * Set the value for each cell 
	 * that contains a possible value than can only be in that cell.
	 */
	private void selectUniquePossibilities() {
		for (Cell cell: cells) {
			if (cell.isKnown()) {
				continue;
			}
			for (Value value:cell.getPossibleValues()) {
				boolean valueIsUnique = true;
				for(Cell otherCell:cells) {
					if (otherCell == cell) {
						continue;
					}
					if (otherCell.getPossibleValues().contains(value)) {
						valueIsUnique = false;
						break;
					}
				}
				if (valueIsUnique) {
					cell.setValue(value);
					break;
				}
			}
			
		}
	}
	
	/**
	 * Remove the values of the known cells from the posibilities of the unknown cells
	 * @throws NoMoreValuesException if a cell has no possible values
	 */
	private void removeDuplicates() throws NoMoreValuesException {
		for (Cell cell: cells) {
			if (cell.isKnown()) {
				for (Cell otherCell: cells) {
					if (cell == otherCell){
						continue;
					}
					otherCell.removePossibleValue(cell.getValue());
				}
			}
		}
	}
	
	/**
	 * Remove groups. 
	 * If 2 cells contain only the same 2 possibilities, then the other cells 
	 * will not contain these 2 possibilities.
	 * @throws NoMoreValuesException if a cell has no possible values
	 */
	private void removeGroups() throws NoMoreValuesException {
		for (Cell cell:cells) {
			if (!cell.isKnown() &&
				 cell.getNrOfPossibilities()==2) {
				Cell matchCell = getCorrespondingGroup(cell);
				if (matchCell != null) {
					for (Cell otherCell: cells) {
						if (cell == otherCell ||
							matchCell == otherCell){
							continue;
						}
						for (Value value: cell.getPossibleValues()) {
							otherCell.removePossibleValue(value);
						}
					}
				}
			}
		}
	}
	
	/**
	 * Finds a cell for which the possibilities are equal to the given cell
	 * @param currentCell the cell
	 * @return
	 */
	private Cell getCorrespondingGroup(Cell currentCell) {
		for (Cell cell:cells) {
			if (cell == currentCell) {
				continue;
			}
			if (cell.arePossiblitiesEqual(currentCell)) {
				return cell;
			}
		}
		return null;
	}
	
	/**
	 * Get the number of possibilities for the group
	 * @return the total number of possibilities
	 */
	public BigInteger getNrOfPossibilities() {
		BigInteger nrOfPossibilities = BigInteger.ONE;
		for (Cell cell: cells) {
			long cellNr = cell.getNrOfPossibilities();
			nrOfPossibilities = nrOfPossibilities.multiply(BigInteger.valueOf(cellNr));
			if (cellNr == 0) {
				System.out.println("Cell "+cell.getRow()+"x"+cell.getColumn()+" has 0 possible values "+cell);
			}
		}
		return nrOfPossibilities;
	}
	
	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer();
		for (Cell cell:cells) {
			sb.append(cell); 
			sb.append('|');
		}
		return sb.toString();
	}
}
